import geopandas as gpd
import numpy as np
import pandas as pd

import justclust.paths.paths as paths

my_paths = paths.define_paths(city = 'bolzano')
# %%

# columns selected for the columns of the outputs
sez_cols = [
    "SEZ", "geometry",
    "SEZ2011", "COD_REG", "COD_ISTAT", "PRO_COM"
    ]

id_col = ['SEZ']

#----    Air Quality    ----

# Define available files
prisk_scr_shp = my_paths.rawdata_dir / "Air quality" / "street_canyon_risk.shp"

# Define columns and other constant values
prisk_raw_cols = ["Area_r1", "Area_r2", "Area_r3", "Area_r4", "Area_r6", "Area_r9"]
prisk_nor_cols = [
    "P_Area_r1",
    "P_Area_r2",
    "P_Area_r3",
    "P_Area_r4",
    "P_Area_r6",
    "P_Area_r9",
]

#----    Carbon    ----

# Define available files
co2_ab_shp = my_paths.rawdata_dir / "Carbon" / "c_absorption.shp"
co2_em_shp = my_paths.rawdata_dir / "Carbon" / "c_emission.shp"

# Define columns and other constant values
# Absorption
co2_raw_cols = ["CO2"]
co2_nor_cols = ["kgCO2/m²"]

# Emmissions
ems = ["gasolio_em", "gpl_em", "ee_em", "metano_em", "legno_em", "tot_em", "c_ee_em"]
co2_em_raw_cols = ["finale"]
co2_em_nor_cols = ["finale/m²"]

#----    FFH    ----

# Define available files
ffh_pa_shp = my_paths.rawdata_dir / "FFH" / "protected_areas.shp"

# Define columns and other constant values
ffh_raw_cols = ["F_Area_m2"]
ffh_nor_cols = ["P_Prot_are"]

#----    Spatial    ----

# Define available files
spt_ag_shp = my_paths.rawdata_dir / "Spatial" / "accessibility_green_areas.shp"
spt_wk_shp = my_paths.rawdata_dir / "Spatial" / "walkability.shp"

# Define columns and other constant values
spt_ag_raw_cols = ["Area_300m_", "Area_750m_"]
spt_ag_nor_cols = ["P_area_300", "P_area_750"]
spt_wk_nor_cols = ["MEAN_Resul"]

#----    Temporal    ----

# Define available files
tem_ss_shp = my_paths.rawdata_dir / "Temporal" / "soil_sealing.shp"

# Define columns and other constant values
tem_raw_cols = ["Area_diff"]
tem_nor_cols = ["P_Area_dif"]

#----    Thermal    ----

# Define available files
thm_sh_shp = my_paths.rawdata_dir / "Thermal" / "suhi.shp"

# Define columns and other constant values
thm_nor_cols = ["mean_norm"]

#----    Socio    ----

# Define available files
socio_gpkg = my_paths.rawdata_dir / "Socio" / "bz_pop_2021.gpkg"

# Define columns and other constant values
socio_abs_cols = [
    "Maschi",
    "Femmine",
    "Totale popolazione",
    "0-14",
    "15-64",
    "65 e oltre",
    "indice di vecchiaia",
    "0-17",
    "Stranieri",
    "Media componenti",
    "Totale famiglie con figli",
    "Totale famiglie senza figli",
    "Totale famiglie",
    "Single",
]
socio_perc_cols = [
    "%65+",
    "%minori",
    "%stranieri",
    "%con figli",
    "%senza figli",
    "%Single",
]


# %%
#----    SELECT COLUMNS TO BE USED    ----

cols_dict = {
    # Air quality
    'prisk':{
        "P_Area_r1": "Area very-low P-risk [%]",
        "P_Area_r2": "Area low P-risk [%]",
        "P_Area_r3": "Area medium-low P-risk [%]",
        "P_Area_r4": "Area medium-high P-risk [%]",
        "P_Area_r6": "Area high P-risk [%]",
        "P_Area_r9": "Area very-high P-risk [%]"
        },
    # Carbon
    'carbon':{
        "finale/m²": "Carbon emission building [ton CO2/m²]",
        "kgCO2/m²": "Carbon absorption vegetation [kg CO2/m²]"
        },
    # FFH
    'ffh':{"P_Prot_are": "Protected area [%]"},
    # Spatial
    'spatial':{
        "P_area_300": "Accessibility public gardens (<5 min) [%]",
        "MEAN_Resul": "Walkability Index"
        },
    # Temporal
    'temporal':{"P_Area_dif":"Soil sealing between 2022 and 2018 [%]"},
    # Thermal
    'thermal':{"mean_norm": "Surface urban heat island"},
    # Socio
    'socio':{
        "indice di vecchiaia": "Age Index",
        "%stranieri": "Foreign population [%]",
        "%con figli": "Families with children [%]",
        # "%senza figli": "Families without children [%]",
        "%Single": "Families with one component [%]"
        }
}

# Get only selected columns old-names 
cols = [col for value in cols_dict.values() for col in list(value.keys())]

# define weights columns
wcols = np.array(
    [weight for value in cols_dict.values() for weight in [1/len(value)] * len(value)],
    dtype=float,
)

# define nice text label to not expose the internal nomenclature
conv = {k:v for values in cols_dict.values() for k,v in values.items()}

# define subset of columns according to categories

selected_col = {
    'col_prisk': list(cols_dict.get('prisk').values()),
    'col_carbon':list(cols_dict.get('carbon').values()),
    'col_char':\
        list(cols_dict.get('ffh').values())+\
        list(cols_dict.get('spatial').values())+\
        list(cols_dict.get('temporal').values())+\
        list(cols_dict.get('thermal').values()),
    'col_socio':list(cols_dict.get('socio').values())
    }


# %%
# Define function to read files


def read_data():
    prisk_scr = gpd.read_file(prisk_scr_shp)
    co2_ab = gpd.read_file(co2_ab_shp)
    co2_ab["SEZ"] = co2_ab["SEZ"].astype(int)
    co2_ab["kgCO2/m²"] = co2_ab["CO2"] / co2_ab.area
    co2_em = gpd.read_file(co2_em_shp)
    co2_em["finale/m²"] = co2_em["finale"] / co2_em.area
    ffh_pa = gpd.read_file(ffh_pa_shp)
    spt_ag = gpd.read_file(spt_ag_shp)
    spt_wk = gpd.read_file(spt_wk_shp)
    tem_ss = gpd.read_file(tem_ss_shp)
    thm_sh = gpd.read_file(thm_sh_shp)

    socio_pk = gpd.read_file(socio_gpkg, driver="GPKG")
    # fix column
    res = []
    for val in socio_pk["indice di vecchiaia"]:
        try:
            v = float(val.strip().replace(",", "."))
        except Exception:
            v = np.nan
        res.append(v)
    socio_pk["indice di vecchiaia"] = np.array(res, dtype=float)

    raw = pd.concat(
        [
            co2_ab.loc[:, sez_cols],
            prisk_scr.loc[:, prisk_raw_cols + prisk_nor_cols],
            co2_ab.loc[:, co2_raw_cols + co2_nor_cols],
            co2_em.loc[:, co2_em_raw_cols + co2_em_nor_cols + ems],
            ffh_pa.loc[:, ffh_raw_cols + ffh_nor_cols],
            spt_ag.loc[:, spt_ag_raw_cols + spt_ag_nor_cols],
            spt_wk.loc[:, spt_wk_nor_cols],
            tem_ss.loc[:, tem_raw_cols + tem_nor_cols],
            thm_sh.loc[:, thm_nor_cols],
            socio_pk.loc[:, socio_abs_cols + socio_perc_cols],
        ],
        axis="columns",
    )
    raw.index = co2_ab["SEZ"]
    raw.index.name = None

    # Remove error observation
    mask = raw.index == 8888888
    raw = raw[~ mask]

    return raw
