from hashlib import shake_128
from itertools import product
from typing import Any, Callable, Dict, List, Optional, Tuple

import pandas as pd
from joblib import Parallel, delayed
from sklearn import decomposition as dec
from sklearn import preprocessing as pre
from tqdm import tqdm

from justclust.hopkins import hopkins

# from pathlib import Path


class ApplyW:
    def __init__(self, weights):
        self.weights = weights

    def fit_transform(self, X, *args, **kwargs):
        return X * self.weights

    def __repr__(self):
        return f"ApplyW(weights={self.weights!r})"


def get_preprocs(selected=None, apply_weight=None):
    scalers = [
        None,
        pre.MinMaxScaler(),
        pre.MaxAbsScaler(),
        pre.StandardScaler(with_mean=True, with_std=True),
        pre.Normalizer(norm="l2"),
        pre.Normalizer(norm="l1"),
        pre.Normalizer(norm="max"), # create sparse matrixes that lead to error in fast ica https://github.com/scikit-learn/scikit-learn/issues/2089#issuecomment-1126183560
        pre.QuantileTransformer(
            n_quantiles=1000,
            output_distribution="uniform",
            ignore_implicit_zeros=False,
            subsample=10000,
            random_state=2023,
        ),
        pre.QuantileTransformer(
            n_quantiles=1000,
            output_distribution="normal",
            ignore_implicit_zeros=False,
            subsample=10000,
            random_state=2023,
        ),
        pre.PowerTransformer(method="yeo-johnson"),
        # The Box-Cox transformation can only be applied to strictly positive data
        # pre.PowerTransformer(method="box-cox"),
    ] + [
        pre.RobustScaler(
            with_centering=True, with_scaling=True, quantile_range=(qmin, qmax)
        )
        for qmin, qmax in [
            (0, 100),
            (1, 99),
            (2, 98),
            (5, 95),
            (10, 90),
            (20, 80),
            (25, 75),
        ]
    ]

    # if selected is not None:
    #     scalers.append(pre.minmax_scale(selected))

    ap_weights = (
        [
            None,
        ]
        if apply_weight is None
        else [None, apply_weight]
    )

    decomps = [
        None,
        dec.PCA(random_state = 2023),
        dec.PCA(n_components="mle", svd_solver="full", random_state = 2023),
        dec.PCA(svd_solver="randomized", whiten=True, random_state = 2023),
        dec.FastICA(whiten="warn", max_iter= 1000,random_state = 2023),
        dec.FastICA(whiten="unit-variance", max_iter= 1000, random_state = 2023),
        dec.FactorAnalysis(max_iter=1000, random_state = 2023),
        dec.DictionaryLearning(fit_algorithm="lars", alpha=0.1, n_jobs=None, 
                               max_iter = 2000, random_state = 2023),
        dec.DictionaryLearning(fit_algorithm="cd", alpha=0.1, n_jobs=None, 
                               max_iter = 2000, random_state = 2023),
    ]
    preprocs = {
        f"s:{s}|w:{'true' if w else 'false'}|d:{d}": (s, w, d)
        for s, w, d in product(scalers, ap_weights, decomps)
    }
    return preprocs


def preprocs_worker(
    data: pd.DataFrame, plabel: str, palgs: List[Optional[Callable]]
) -> Tuple[str, Dict[str, str | float]]:
    """Excute the single model and compute the main scores

    Parameters
    ----------
    data : pd.DataFrame
        Selected data
    plabel: str
        Name of the algorithm in use.
    palgs: Dict[str, Any]
        Dictionary with the main options and parameters

    Returns
    -------
    Tuple[str, List[int], List[Any], pd.DataFrame]
        * the first value is the string that is used as column name
        * the result of the cluster with the assigned labels for
        each row of the transformed-data
        * list containing the scores and main features of the cluster
        * DataFrame containing the feature importance of each cluster
    """
    df = data.copy()
    # print(plabel)
    s, w, d = palgs

    # print(f"    - scaler: {s!r}")
    if s is not None:
        df = s.fit_transform(df)
        sn = s.__class__.__name__
    else:
        sn = "None"

    # print(f"    - weight: {w!r}")
    if w is not None:
        df = w.fit_transform(df)
        wn = w.__class__.__name__
    else:
        wn = "None"

    # print(f"    - decomposer: {d!r}")
    if d is not None:
        df = d.fit_transform(df)
        dn = d.__class__.__name__
    else:
        dn = "None"

    # check sample is no greater than df size
    sampling_size = min(df.shape[0], 150)

    hop = hopkins(df, sampling_size=sampling_size)

    long_label = f"s:{s!r}|w:{w!r}|d:{d!r}"
    pre_id = shake_128(long_label.encode()).hexdigest(4)

    return (
        plabel,
        dict(
            key=plabel,
            pre_id=pre_id,
            short_label=f"s:{sn}|w:{wn}|d:{dn}",
            long_label=long_label,
            scaler_class=sn,
            scaler_details=f"{s!r}",
            weight_class=wn,
            decomposer_class=dn,
            decomposer_details=f"{d!r}",
            hopkins=hop,
        ),
    )


def explore_preprocs(
    data: pd.DataFrame,
    preprocs: Dict[str, Dict[str, Any]] = None,
    n_jobs: int = -1,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Execute all the pre-process options and see
    if the tendency to cluster improve or not.

    Parameters
    ----------
    data : pd.DataFrame
        Data to be used for the clustering, is a DataFrame with
        the selected features.
    preprocs : Dict[str, Dict[str, Any]]
        A dictionary of dicionary with all the pre-processing options
    n_jobs: int, default: -1
        Number of parallel jobs to be used to explore the
        algorithms. A negative value mean that all the cores
        available will be used.

    Returns
    -------
    * the DataFrame with the main score and characteristics
    of the clusters
    * the DataFrame with the feature importance
    """
    if preprocs is None:
        preprocs = get_preprocs()

    num_opts = len(preprocs.keys())
    print(f"Exploring and testing {num_opts} of pre-processing options")

    pll = Parallel(n_jobs=n_jobs, verbose=0)
    tasks = [(plabel, palg) for plabel, palg in preprocs.items()]

    res = pll(
        delayed(preprocs_worker)(data, plabel, palg) for plabel, palg in tqdm(tasks)
    )

    index = [idx for idx, _ in res]
    records = [rec for _, rec in res]
    scores = pd.DataFrame.from_records(records, index=index)
    # scores.set_index("pre_id", inplace=True)
    scores.sort_values("hopkins", ascending=True, inplace=True)

    return scores
